# Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import logging
import time
from typing import Any

from google.protobuf.empty_pb2 import Empty
from google.protobuf.text_format import MessageToString
from th2_check2_recon.configuration import CrawlerConnectionConfiguration
from th2_check2_recon.reconcommon import ReconMessage, ReconMessageUtils
from th2_common.schema.message.message_listener import MessageListener
from th2_common_utils import message_to_dict
from th2_grpc_common.common_pb2 import EventID, MessageBatch
from th2_grpc_crawler_data_processor.crawler_data_processor_pb2 import DataProcessorInfo, EventResponse, \
    MessageResponse, Status
from th2_grpc_crawler_data_processor.crawler_data_processor_pb2_grpc import DataProcessorServicer

logger = logging.getLogger(__name__)


class AbstractHandler(MessageListener, ABC):
    def __init__(self, rule: Any) -> None:
        self._rule = rule

    @abstractmethod
    def handler(self, attributes: tuple, batch: Any) -> Any:
        pass


class MessageHandler(AbstractHandler):
    def handler(self, attributes: tuple, batch: MessageBatch) -> None:
        try:
            for proto_message in batch.messages:
                data = message_to_dict(proto_message)
                message = ReconMessage(proto_message=data)

                process_timer = self._rule.RULE_PROCESSING_TIME
                start_time = time.time()

                self._rule.clear_out_of_timeout(message.timestamp_ns)
                self._rule.check_no_match_within_timeout(actual_timestamp=message.timestamp,
                                                         actual_timestamp_ns=message.timestamp_ns)
                if self._rule.has_been_grouped(message, attributes):
                    self._rule.process(message, attributes)

                process_timer.observe(time.time() - start_time)

                logger.debug("Processed '%s' id='%s'" %
                             (data['metadata']['message_type'], ReconMessageUtils.str_message_id(data)))

            logger.debug("Cache size '%s': %s." % (self._rule.get_name(), self._rule.log_groups_size()))

        except Exception as e:
            logger.exception(f'Rule: {self._rule.get_name()}. '
                             f'An error occurred while processing the received message: {e}\n'
                             f'Message: {MessageToString(batch, as_one_line=True)}')


class GRPCHandler(DataProcessorServicer):

    def __init__(self,
                 rules: list,
                 crawler_connection_configuration:
                 CrawlerConnectionConfiguration,
                 root_event_id: EventID) -> None:
        self._rules: list = rules
        self._crawler_connection_configuration = crawler_connection_configuration
        self._recon_root_event_id: EventID = root_event_id

    def CrawlerConnect(self, request: Any, context: Any) -> DataProcessorInfo:
        logger.debug('CrawlerId %s connected' % request.id.name)
        return DataProcessorInfo(name=self._crawler_connection_configuration.name,
                                 version=self._crawler_connection_configuration.version)

    def IntervalStart(self, request: Any, context: Any) -> Empty:
        logger.debug('Interval set from %s to %s' % (request.start_time, request.end_time))
        return Empty()

    def SendEvent(self, request: Any, context: Any) -> EventResponse:
        logger.debug('CrawlerID %s sent events %s' %
                     (request.id.name, MessageToString(request.event_data, as_one_line=True)))
        return EventResponse(id=self._recon_root_event_id, status=Status(handshake_required=False))

    def SendMessage(self, request: Any, context: Any) -> MessageResponse:
        try:
            logger.debug('CrawlerID %s sent %s messages' % (request.id.name, len(request.message_data)))
            messages = [data.message for data in request.message_data]
            for proto_message in messages:
                data = message_to_dict(proto_message)
                for rule in self._rules:
                    message = ReconMessage(proto_message=data)
                    process_timer = rule.RULE_PROCESSING_TIME
                    start_time = time.time()
                    try:
                        rule.clear_out_of_timeout(message.timestamp_ns)
                        rule.check_no_match_within_timeout(actual_timestamp=message.timestamp,
                                                           actual_timestamp_ns=message.timestamp_ns)
                        if rule.has_been_grouped(message, ()):
                            rule.process(message, ())

                    except Exception as e:
                        logger.exception(
                            f'Rule: {rule.get_name()}. '
                            f'An error occurred while processing the message: {e}\n'
                            f'Message: {MessageToString(proto_message, as_one_line=True)}')
                    finally:
                        process_timer.observe(time.time() - start_time)
                logger.debug("Processed '%s' id='%s'"
                             % (data['metadata']['message_type'], ReconMessageUtils.str_message_id(data)))
            return MessageResponse(ids=[msg.metadata.id for msg in messages], status=Status(handshake_required=False))
        except Exception as e:
            logger.exception(f'SendMessage request failed: {e}')
            return MessageResponse(ids=[], status=Status(handshake_required=True))



# Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Dict, Optional, Union


class RuleConfiguration:

    def __init__(self,
                 name: str,
                 enabled: str,
                 match_timeout: str,
                 match_timeout_offset_ns: str,
                 match_all: Optional[str] = None,
                 autoremove_timeout: Optional[str] = None,
                 configuration: Optional[str] = None) -> None:
        self.name = str(name)
        self.enabled = enabled.lower() == 'true'
        self.match_timeout = int(match_timeout)
        self.match_timeout_offset_ns = int(match_timeout_offset_ns)
        self.match_all = match_all is not None and match_all.lower() == 'true'
        self.autoremove_timeout: Optional[Union[int, datetime.datetime]]
        if autoremove_timeout is not None:
            try:
                timeout_int = int(autoremove_timeout)
                self.autoremove_timeout = timeout_int
            except ValueError:
                timeout = datetime.datetime.strptime(autoremove_timeout, '%H:%M')
                self.autoremove_timeout = timeout.combine(datetime.datetime.now().date(), timeout.time())
        else:
            self.autoremove_timeout = None
        self.configuration = configuration


class CrawlerConnectionConfiguration:

    def __init__(self, name: str = 'Recon Data Processor', version: str = '1.0.0') -> None:
        self.name = name
        self.version = version


class ReconConfiguration:
    def __init__(self,
                 recon_name: str,
                 cache_size: int,
                 event_batch_max_size: int,
                 event_batch_send_interval: int,
                 rules_package_path: str,
                 rules: list,
                 crawler_connection_configuration: Optional[Dict[str, str]] = None,
                 configuration: Optional[str] = None) -> None:
        self.recon_name = recon_name
        self.cache_size = int(cache_size)
        self.event_batch_max_size = int(event_batch_max_size)
        self.event_batch_send_interval = int(event_batch_send_interval)
        self.rules_package_path = rules_package_path
        self.rules = [RuleConfiguration(**rule) for rule in rules]

        if crawler_connection_configuration is None:
            self.crawler_connection_configuration = CrawlerConnectionConfiguration()
        else:
            self.crawler_connection_configuration = CrawlerConnectionConfiguration(**crawler_connection_configuration)

        self.configuration = configuration



# Copyright 2020-2020 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import collections
import datetime
import enum
import itertools
import logging
from threading import Lock, Timer
from typing import Any, Dict, Iterator, List, Optional, Union

import sortedcollections
from th2_check2_recon.common import EventType, MessageComponent, \
    VerificationComponent
from th2_check2_recon.reconcommon import MessageGroupDescription, ReconMessage
from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common_utils import dict_to_message
from th2_common_utils import event_utils
from th2_grpc_common.common_pb2 import ConnectionID, Direction, Event, EventBatch, EventID, EventStatus, \
    Message, MessageID
from th2_grpc_util.util_pb2 import CompareMessageVsMessageRequest, CompareMessageVsMessageResult, \
    CompareMessageVsMessageTask, ComparisonSettings

logger = logging.getLogger(__name__)


class AbstractService(ABC):

    @abstractmethod
    def stop(self) -> None:
        pass


class EventSender:

    class RepeatTimer(Timer):

        def run(self) -> None:
            while not self.finished.wait(self.interval):  # type: ignore
                self.function(*self.args, **self.kwargs)  # type: ignore

    def __init__(self,
                 event_router: EventBatchRouter,
                 event_batch_max_size: int,
                 event_batch_sending_interval: int):
        self.event_router = event_router
        self.max_batch_size = event_batch_max_size
        self.sending_interval = event_batch_sending_interval

        self._event_batches_to_send: Dict[str, EventBatch] = {}
        self._lock: Lock = Lock()
        self._timer = self.create_timer()

    def create_timer(self) -> Timer:
        timer = EventSender.RepeatTimer(self.sending_interval, self._timer_handle)
        timer.start()
        return timer

    def _timer_handle(self) -> None:
        with self._lock:
            for event_batch in self._event_batches_to_send.values():
                self.send_event_batch(event_batch)
            self._event_batches_to_send.clear()

    def send_event(self, event: Event) -> None:
        try:
            self.event_router.send(EventBatch(events=[event]))
        except Exception as e:
            logger.exception(f'Could not send event: {e}')

    def send_event_batch(self, event_batch: EventBatch) -> None:
        try:
            self.event_router.send(event_batch)
        except Exception as e:
            logger.exception(f'Could not send event batch: {e}')

    def put_event_in_batch(self, event: Event) -> None:
        with self._lock:
            parent_id = event.parent_id
            if parent_id.id in self._event_batches_to_send:
                event_batch = self._event_batches_to_send[parent_id.id]
            else:
                event_batch = EventBatch(parent_event_id=parent_id)
                self._event_batches_to_send[parent_id.id] = event_batch

            event_batch.events.append(event)
            if len(event_batch.events) == self.max_batch_size:
                del self._event_batches_to_send[parent_id.id]
                self.send_event_batch(event_batch)

    def stop(self) -> None:
        with self._lock:
            self._timer.cancel()
            for event_batch in self._event_batches_to_send.values():
                self.send_event_batch(event_batch)
            self._event_batches_to_send.clear()


class EventStore(AbstractService):

    class EventPack(str, enum.Enum):
        FAILED_MATCHES = 'Failed matches'
        PASSED_MATCHES = 'Passed matches'
        OUT_OF_TIMEOUT_MATCHES = 'Out of timeout matches'
        NO_MATCH_WITHIN_TIMEOUT = 'No match within timeout'
        IGNORED_NO_MATCH_WITHIN_TIMEOUT = 'Ignored no match within timeout'
        REMOVED_MESSAGES = 'Removed messages'
        ERRORS = 'Errors'

    class EventCollection:

        def __init__(self, parent_event_id: EventID):
            self.parent_event_id: EventID = parent_event_id
            self.event_pack_ids: Dict[str, Optional[EventID]] = {
                EventStore.EventPack.FAILED_MATCHES: None,
                EventStore.EventPack.IGNORED_NO_MATCH_WITHIN_TIMEOUT: None,
                EventStore.EventPack.PASSED_MATCHES: None,
                EventStore.EventPack.OUT_OF_TIMEOUT_MATCHES: None,
                EventStore.EventPack.NO_MATCH_WITHIN_TIMEOUT: None,
                EventStore.EventPack.REMOVED_MESSAGES: None,
                EventStore.EventPack.ERRORS: None
            }

        def get_event_pack_id(self, event_pack_name: 'EventStore.EventPack') -> Optional[EventID]:
            return self.event_pack_ids[event_pack_name]

        def create_pack_event(self, event_pack_name: 'EventStore.EventPack') -> Event:
            event: Event = event_utils.create_event(name=event_pack_name,
                                                    parent_id=self.parent_event_id,
                                                    event_type=EventType.STATUS)
            self.event_pack_ids[event_pack_name] = event.id
            return event

    def __init__(self,
                 recon_name: str,
                 event_router: EventBatchRouter,
                 event_batch_max_size: int,
                 event_batch_sending_interval: int) -> None:
        self.event_sender: EventSender = EventSender(event_router=event_router,
                                                     event_batch_max_size=event_batch_max_size,
                                                     event_batch_sending_interval=event_batch_sending_interval)
        self.recon_event_id = self.put_recon_event(recon_name=recon_name)
        self.shared_groups_root_event_id = self.put_shared_groups_root_event()
        self.shared_group_events: Dict[str, EventID] = {}
        self.event_tree: Dict[str, EventStore.EventCollection] = {}

    def put_recon_event(self, recon_name: str) -> EventID:
        recon_event: Event = event_utils.create_event(name=f'Recon: {recon_name}', event_type=EventType.ROOT)
        self.event_sender.put_event_in_batch(event=recon_event)
        return recon_event.id

    def put_shared_groups_root_event(self) -> EventID:
        shared_root_event_id: Event = event_utils.create_event(parent_id=self.recon_event_id,
                                                               name='Shared groups',
                                                               event_type=EventType.EVENT)
        self.event_sender.put_event_in_batch(event=shared_root_event_id)
        return shared_root_event_id.id

    def put_shared_group_event(self, shared_group_name: str) -> EventID:
        shared_group_event: Event = event_utils.create_event(parent_id=self.shared_groups_root_event_id,
                                                             name=f'Group: {shared_group_name}',
                                                             event_type=EventType.EVENT)
        self.event_tree[shared_group_event.id.id] = EventStore.EventCollection(shared_group_event.id)
        self.event_sender.put_event_in_batch(event=shared_group_event)
        return shared_group_event.id

    def put_rule_event(self, rule_name: str, rule_description: str) -> EventID:
        rule_event: Event = event_utils.create_event(name=rule_name,
                                                     parent_id=self.recon_event_id,
                                                     body=event_utils.create_event_body(
                                                         MessageComponent(message=rule_description)),
                                                     event_type=EventType.RULE)
        self.event_tree[rule_event.id.id] = EventStore.EventCollection(rule_event.id)
        self.event_sender.put_event_in_batch(event=rule_event)
        return rule_event.id

    def put_event_in_group_pack(self,
                                event: Event,
                                rule_event_id: EventID,
                                event_pack_name: EventPack,
                                shared_group_name: Optional[str] = None) -> None:
        if shared_group_name is not None:
            if shared_group_name in self.shared_group_events:
                event_id = self.shared_group_events[shared_group_name]
            else:
                event_id = self.put_shared_group_event(shared_group_name)
                self.shared_group_events[shared_group_name] = event_id
        else:
            event_id = rule_event_id

        event_pack_id = self.event_tree[event_id.id].get_event_pack_id(event_pack_name)
        if event_pack_id is None:
            pack_event = self.event_tree[event_id.id].create_pack_event(event_pack_name)
            event_pack_id = pack_event.id
            self.event_sender.put_event_in_batch(event=pack_event)

        event.parent_id.CopyFrom(event_pack_id)
        self.event_sender.put_event_in_batch(event=event)

    def create_and_send_single_message_event(self,
                                             recon_message: ReconMessage,
                                             event_name: str,
                                             event_message: str,
                                             event_status: Union[str, int],
                                             rule_event_id: EventID,
                                             event_pack_name: EventPack) -> None:
        event: Event = event_utils.create_event(name=event_name,
                                                body=event_utils.create_event_body(MessageComponent(event_message)),
                                                attached_message_ids=self._get_attached_message_ids(recon_message),
                                                status=event_status,
                                                event_type=EventType.EVENT)
        self.put_event_in_group_pack(event=event,
                                     rule_event_id=rule_event_id,
                                     event_pack_name=event_pack_name,
                                     shared_group_name=recon_message.group_name if recon_message._shared else None)

    def store_no_match_within_timeout_event(self,
                                            rule_event_id: EventID,
                                            recon_message: ReconMessage,
                                            actual_timestamp: datetime.datetime,
                                            timeout: int,
                                            event_pack_name: 'EventStore.EventPack') -> None:
        event_message = f'Timestamp of the last received message: {actual_timestamp}\nTimestamp this message: ' \
                        f'{recon_message.timestamp}\n Timeout: {timeout / 10 ** 9}'
        event_status = EventStatus.SUCCESS \
            if event_pack_name == EventStore.EventPack.IGNORED_NO_MATCH_WITHIN_TIMEOUT \
            else EventStatus.FAILED
        self.create_and_send_single_message_event(recon_message=recon_message,
                                                  event_name=f'{recon_message.all_info}',
                                                  event_message=event_message,
                                                  event_status=event_status,
                                                  rule_event_id=rule_event_id,
                                                  event_pack_name=event_pack_name)

    def store_removed_message_event(self,
                                    rule_event_id: EventID,
                                    recon_message: ReconMessage,
                                    event_message: str) -> None:
        event_message += f'\n Message {"not" if not recon_message._is_matched else ""} matched'
        event_status = EventStatus.SUCCESS if recon_message._is_matched else EventStatus.FAILED
        self.create_and_send_single_message_event(recon_message=recon_message,
                                                  event_name=f'Remove {recon_message.all_info}',
                                                  event_message=event_message,
                                                  event_status=event_status,
                                                  rule_event_id=rule_event_id,
                                                  event_pack_name=EventStore.EventPack.REMOVED_MESSAGES)

    def store_error(self, rule_event_id: EventID, event_name: str,
                    error_message: str, messages: Optional[List[ReconMessage]] = None) -> None:
        body = event_utils.create_event_body(MessageComponent(error_message))
        attached_message_ids = self._get_attached_message_ids(*messages) if messages is not None else []
        event: Event = event_utils.create_event(name=event_name,
                                                status=EventStatus.FAILED,
                                                attached_message_ids=attached_message_ids,
                                                body=body,
                                                event_type=EventType.EVENT)
        self.put_event_in_group_pack(event=event,
                                     rule_event_id=rule_event_id,
                                     event_pack_name=EventStore.EventPack.ERRORS)

    def store_matched_out_of_timeout(self, rule_event_id: EventID, check_event: Event) -> None:
        self.put_event_in_group_pack(event=check_event,
                                     rule_event_id=rule_event_id,
                                     event_pack_name=EventStore.EventPack.OUT_OF_TIMEOUT_MATCHES)

    def store_matched(self, rule_event_id: EventID, check_event: Event) -> None:
        if check_event.status == EventStatus.SUCCESS:
            self.put_event_in_group_pack(event=check_event,
                                         rule_event_id=rule_event_id,
                                         event_pack_name=EventStore.EventPack.PASSED_MATCHES)
        else:
            self.put_event_in_group_pack(event=check_event,
                                         rule_event_id=rule_event_id,
                                         event_pack_name=EventStore.EventPack.FAILED_MATCHES)

    def stop(self) -> None:
        self.event_sender.stop()

    @staticmethod
    def _get_attached_message_ids(*recon_msgs: ReconMessage) -> List[Optional[MessageID]]:
        try:
            return [
                MessageID(
                    connection_id=ConnectionID(
                        session_alias=message.proto_message['metadata']['session_alias'],
                        session_group=message.proto_message['metadata']['session_group']),
                    direction=getattr(Direction, message.proto_message['metadata']['direction']),
                    sequence=message.proto_message['metadata']['sequence'],
                    subsequence=message.proto_message['metadata']['subsequence'])
                for message in recon_msgs
            ]
        except KeyError:
            return []


class MessageComparator(AbstractService):

    def __init__(self, comparator_service) -> None:  # type: ignore
        self.__comparator_service = comparator_service

    def compare(self, expected: Message, actual: Message,
                settings: ComparisonSettings) -> CompareMessageVsMessageResult:
        return self.comparing(expected, actual, settings)

    def comparing(self, expected: Message, actual: Message,
                  settings: ComparisonSettings) -> CompareMessageVsMessageResult:
        request = CompareMessageVsMessageRequest(comparison_tasks=[
            CompareMessageVsMessageTask(first=expected, second=actual, settings=settings)
        ])
        try:
            compare_response = self.__comparator_service.compareMessageVsMessage(request)
            for compare_result in compare_response.comparison_results:
                return compare_result  # type:ignore
        except Exception as e:
            logger.exception(f'Error while comparing: {e}.\nCompareMessageVsMessageRequest: {request}')
        return CompareMessageVsMessageResult()

    def compare_messages(self, messages: List[ReconMessage],
                         ignore_fields: Optional[List[str]] = None) -> Optional[VerificationComponent]:
        if len(messages) != 2:
            logger.exception('The number of messages to compare must be 2.')
            return None
        settings = ComparisonSettings()
        if ignore_fields is not None:
            settings.ignore_fields.extend(ignore_fields)

        compare_result = self.compare(dict_to_message(messages[0].proto_message['fields']),
                                      dict_to_message(messages[1].proto_message['fields']), settings)
        return VerificationComponent(compare_result.comparison_result)

    def stop(self) -> None:
        pass


class Cache(AbstractService):

    __slots__ = 'rule', 'capacity', 'event_store', 'rule_event', '_message_groups'

    def __init__(self, rule: Any, cache_size: int) -> None:
        self.rule = rule
        self.capacity = cache_size
        self.event_store = self.rule.event_store
        self.rule_event_id = self.rule.rule_event_id

        self._message_groups = {}

        for group_name, group_description in self.rule.description_of_groups().items():
            if group_description.shared:
                if group_name not in self.rule.recon.shared_message_groups:
                    self.rule.recon.shared_message_groups[group_name] = \
                        Cache.MessageGroup(group_name=group_name,
                                           group_description=group_description,
                                           capacity=cache_size,
                                           event_store=self.event_store,
                                           parent_event_id=self.rule_event_id)
                self._message_groups[group_name] = self.rule.recon.shared_message_groups[group_name]
            else:
                self._message_groups[group_name] = Cache.MessageGroup(group_name=group_name,
                                                                      group_description=group_description,
                                                                      capacity=cache_size,
                                                                      event_store=self.event_store,
                                                                      parent_event_id=self.rule_event_id)

        has_multi = any(group.group_description.multi for group in self._message_groups.values())
        if has_multi:
            for group in self.message_groups.values():
                if group.group_description.single:
                    group.is_cleanable = False

    @property
    def message_groups(self) -> dict:
        return self._message_groups

    def stop(self) -> None:
        for group in self.message_groups.values():
            group.clear()

    class MessageGroup:

        __slots__ = (
            'name', 'capacity', 'size', 'group_description', 'event_store', 'parent_event_id', 'is_cleanable',
            'data', 'hash_by_sorted_timestamp'
        )

        def __init__(self,
                     group_name: str,
                     capacity: int,
                     group_description: MessageGroupDescription,
                     event_store: EventStore,
                     parent_event_id: EventID) -> None:
            self.name = group_name
            self.capacity = capacity
            self.size = 0
            self.group_description = group_description
            self.event_store = event_store
            self.parent_event_id: EventID = parent_event_id

            self.is_cleanable = True
            self.data: Dict[int, List[ReconMessage]] = collections.defaultdict(list)
            self.hash_by_sorted_timestamp: Dict[int, List[int]] = sortedcollections.SortedDict()

        def put(self, message: ReconMessage) -> None:
            if self.size < self.capacity and message.hash is not None:
                if message.hash in self.data and self.group_description.single:
                    if self.group_description.shared:
                        return
                    cause = f'The message was deleted because a new one came with the same hash "{message.hash}" ' \
                            f'in message group "{self.name}"'
                    self.remove(message.hash, cause)

                self.data[message.hash].append(message)
                self.hash_by_sorted_timestamp.setdefault(message.timestamp_ns, []).append(message.hash)
                self.size += 1
            else:
                timestamp_for_remove = next(iter(self.hash_by_sorted_timestamp.keys()))
                hash_for_remove = self.hash_by_sorted_timestamp[timestamp_for_remove][0]
                cause = f'The message was deleted because there was no free space in the message group "{self.name}"'
                self.remove(hash_for_remove, cause, remove_all=False)
                self.put(message)

        def check_no_match_within_timeout(self,
                                          actual_timestamp: datetime.datetime,
                                          actual_timestamp_ns: int,
                                          timeout: int) -> None:
            lower_bound_timestamp = actual_timestamp_ns - timeout
            for timestamp, hashes in self.hash_by_sorted_timestamp.items():
                if timestamp < lower_bound_timestamp:
                    old_hash = hashes[0]
                    for recon_message in self.data[old_hash]:
                        if not recon_message._is_matched and not recon_message._was_checked_no_match_within_timeout:
                            recon_message._was_checked_no_match_within_timeout = True
                            if self.group_description.ignore_no_match:
                                self.event_store.store_no_match_within_timeout_event(
                                    rule_event_id=self.parent_event_id,
                                    recon_message=recon_message,
                                    actual_timestamp=actual_timestamp,
                                    timeout=timeout,
                                    event_pack_name=EventStore.EventPack.IGNORED_NO_MATCH_WITHIN_TIMEOUT)
                                self.remove(hash_of_message=old_hash)
                            else:
                                self.event_store.store_no_match_within_timeout_event(
                                    rule_event_id=self.parent_event_id,
                                    recon_message=recon_message,
                                    actual_timestamp=actual_timestamp,
                                    timeout=timeout,
                                    event_pack_name=EventStore.EventPack.NO_MATCH_WITHIN_TIMEOUT)
                else:
                    break

        def remove(self, hash_of_message: int, cause: Optional[str] = None, remove_all: bool = True) -> None:
            message_for_remove = None
            if remove_all:
                for message_for_remove in self.data[hash_of_message]:
                    timestamp_for_remove = message_for_remove.timestamp_ns
                    self.hash_by_sorted_timestamp[timestamp_for_remove].remove(hash_of_message)
                    if len(self.hash_by_sorted_timestamp[timestamp_for_remove]) == 0:
                        del self.hash_by_sorted_timestamp[timestamp_for_remove]
                    self.size -= 1
                del self.data[hash_of_message]
            else:
                message_for_remove = self.data[hash_of_message][0]
                timestamp_for_remove = message_for_remove.timestamp_ns

                self.data[hash_of_message].remove(message_for_remove)
                if len(self.data[hash_of_message]) == 0:
                    del self.data[hash_of_message]

                self.hash_by_sorted_timestamp[timestamp_for_remove].remove(hash_of_message)
                if len(self.hash_by_sorted_timestamp[timestamp_for_remove]) == 0:
                    del self.hash_by_sorted_timestamp[timestamp_for_remove]
                self.size -= 1

            if cause is not None:
                self.event_store.store_removed_message_event(rule_event_id=self.parent_event_id,
                                                             recon_message=message_for_remove,
                                                             event_message=cause)

        def clear(self) -> None:
            cause = 'The message was deleted because the Recon stopped'
            while len(self.data) != 0:
                hash_for_remove = next(iter(self.data.keys()))
                while hash_for_remove in self.data and len(self.data[hash_for_remove]) != 0:
                    self.remove(hash_for_remove, cause, remove_all=False)

        def wipe(self) -> None:
            self.data.clear()
            self.hash_by_sorted_timestamp.clear()
            self.size = 0

        def clear_out_of_timeout(self, clean_timestamp: int) -> None:
            hashes_to_removal = []
            for timestamp, hashes in self.hash_by_sorted_timestamp.items():
                if timestamp < clean_timestamp:
                    hashes_to_removal.append(hashes)
                else:
                    break
            for hash_to_remove in itertools.chain.from_iterable(hashes_to_removal):
                self.remove(hash_to_remove, remove_all=False)

        def __iter__(self) -> Iterator[ReconMessage]:
            """Generator that yields all messages in the group"""
            for recon_mgs_lst in self.data.values():
                yield from recon_mgs_lst

        def __contains__(self, hash_of_message: int) -> bool:
            return hash_of_message in self.data



# Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
import datetime
import functools
import logging
import traceback
from typing import Any, Dict, Generator, List, Optional

from prometheus_client import Histogram
from th2_check2_recon.handler import AbstractHandler, MessageHandler
from th2_check2_recon.reconcommon import _get_msg_timestamp, MessageGroupDescription, ReconMessage
from th2_check2_recon.services import Cache, EventStore, MessageComparator
from th2_common.schema.metrics import common_metrics
from th2_grpc_common.common_pb2 import Event, EventID


logger = logging.getLogger(__name__)


class Rule:

    def __init__(self,
                 recon: Any,
                 cache_size: int,
                 match_timeout: int,
                 match_all: bool,
                 autoremove_timeout: Optional[int],
                 configuration: str,
                 metric_number: int) -> None:
        self.recon = recon
        self.configure(configuration)
        self.name = self.get_name()
        logger.info(f"Rule '{self.name}' is initializing...")

        self.event_store: EventStore = recon.event_store
        self.message_comparator: Optional[MessageComparator] = recon.message_comparator
        self.match_timeout = match_timeout
        self.autoremove_timeout = autoremove_timeout
        self.match_all = match_all

        self.reprocess_queue: List[ReconMessage] = []

        self.rule_event_id: EventID = self.event_store.put_rule_event(rule_name=self.name,
                                                                      rule_description=self.get_description())

        self.__cache = Cache(self, cache_size)

        self.RULE_PROCESSING_TIME = Histogram(f'th2_recon_rule_{metric_number}_processing_time',
                                              'Time of the message processing with a rule',
                                              buckets=common_metrics.DEFAULT_BUCKETS)

        logger.info(f"Rule '{self.name}' was initialized")

    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_description(self) -> str:
        pass

    @abstractmethod
    def get_attributes(self) -> List[str]:
        pass

    @abstractmethod
    def description_of_groups(self) -> Dict[str, MessageGroupDescription]:
        """
            Returns a dictionary whose key is group_name and
            value is MessageGroupDescription.
            MessageGroupDescription have boolean fields single, multi,
            shared and ignore_no_match.
        """
        pass

    @property
    def common_configuration(self) -> Optional[str]:
        """Returns recon-level rule configuration if it was set, else - None.
        """
        return self.recon._config.configuration  # type: ignore

    @abstractmethod
    def group(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def hash(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def check(self, messages: List[ReconMessage], attributes: tuple, *args: Any, **kwargs: Any) -> Optional[Event]:
        pass

    def configure(self, configuration: str) -> None:
        pass

    def get_listener(self) -> AbstractHandler:
        return MessageHandler(self)

    def put_shared_message(self, message: ReconMessage) -> None:
        # message = copy.deepcopy(message)
        self.recon.put_shared_message(message)

    def has_been_grouped(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> bool:
        if message.group_name is None:
            self.__group_and_store_event(message, attributes, *args, **kwargs)
            if message.group_name is None:
                logger.debug("RULE '%s': Ignored %s due to unset group_id" % (self.name, message.all_info))
                return False
        return True

    def process(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> None:
        self.__hash_and_store_event(message, attributes, *args, **kwargs)
        if message.hash is None:
            logger.debug("RULE '%s': Ignored %s due to unset hash" % (self.name, message.all_info))
            return

        if message.group_name not in self.__cache.message_groups:
            raise Exception(f'"group" method set incorrect groups.\n - message: {message.all_info}\n'
                            f' - available groups: {self.description_of_groups()}\n'
                            f' - message.group_id: {message.group_name}')
        main_group = self.__cache.message_groups[message.group_name]
        logger.debug("RULE '%s': Received %s" % (self.name, message.all_info))

        for match in self.find_matched_messages(message, match_whole_list=self.match_all):
            # Will return None if some groups did not contain the message.
            if match is None:
                main_group.put(message)
                return
            else:
                self.__check_and_store_event(match, attributes, *args, **kwargs)

        if self.match_all:
            main_group.put(message)

        for group_id, group in self.__cache.message_groups.items():
            if group_id != message.group_name and group.is_cleanable:
                group.remove(message.hash)

        for message in self.reprocess_queue:
            self.process(message, attributes)
        self.reprocess_queue.clear()

    def find_matched_messages(self, message: ReconMessage,
                              match_whole_list: bool = False) -> Generator[Optional[list], ReconMessage, None]:
        if len(self.description_of_groups()) == 1:
            yield [message]
            return

        matched_messages = []
        for group_name, group in self.__cache.message_groups.items():
            if group_name == message.group_name:
                continue
            if message.hash not in group:
                yield None
            matched_messages.append(group.data[message.hash])

        if match_whole_list:
            yield [message] + functools.reduce(lambda inner_list1, inner_list2:
                                               inner_list1 + inner_list2,
                                               matched_messages)
            return

        for match in zip(*matched_messages):
            yield [message] + list(match)

    def queue_for_retransmitting(self, message: ReconMessage) -> None:
        self.reprocess_queue.append(message)

    def clear_cache(self) -> None:
        for group in self.__cache.message_groups.values():
            group.wipe()

    def __group_and_store_event(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> None:
        try:
            self.group(message, attributes, *args, **kwargs)
        except Exception as e:
            logger.exception(f'RULE "{self.name}" - An exception was caught while running "group": {e}')
            self.event_store.store_error(rule_event_id=self.rule_event_id,
                                         event_name='An exception was caught while running "group"',
                                         error_message=traceback.format_exc(),
                                         messages=[message])

    def __hash_and_store_event(self, message: ReconMessage, attributes: tuple, *args: Any, **kwargs: Any) -> None:
        try:
            self.hash(message, attributes, *args, **kwargs)
        except Exception as e:
            logger.exception(f'RULE "{self.name}" - An exception was caught while running "hash": {e}')
            self.event_store.store_error(rule_event_id=self.rule_event_id,
                                         event_name='An exception was caught'
                                                    ' while running "hash"',
                                         error_message=traceback.format_exc(),
                                         messages=[message])

    def __check_and_store_event(self, matched_messages: List[ReconMessage],
                                attributes: tuple, *args: Any, **kwargs: Any) -> None:
        for msg in matched_messages:
            msg._is_matched = True
        try:
            check_event: Optional[Event] = self.check(matched_messages, attributes, *args, **kwargs)
        except Exception as e:
            logger.exception(f"RULE '{self.name}' - An exception was caught while running 'check': {e}")
            self.event_store.store_error(rule_event_id=self.rule_event_id,
                                         event_name='An exception was caught '
                                                    'while running "check"',
                                         error_message=traceback.format_exc(),
                                         messages=matched_messages)
            return

        if check_event is None:
            return

        max_timestamp_msg = max(matched_messages, key=_get_msg_timestamp)
        min_timestamp_msg = min(matched_messages, key=_get_msg_timestamp)

        if max_timestamp_msg.timestamp_ns - min_timestamp_msg.timestamp_ns > self.match_timeout:
            self.event_store.store_matched_out_of_timeout(rule_event_id=self.rule_event_id, check_event=check_event)
        else:
            self.event_store.store_matched(rule_event_id=self.rule_event_id, check_event=check_event)

    def log_groups_size(self) -> str:
        return '[' + ', '.join(f'"{group.name}": {group.size} msg' for group
                               in self.__cache.message_groups.values()) + ']'

    def cache_size(self) -> int:
        return sum(group.size for group in self.__cache.message_groups.values())

    def check_no_match_within_timeout(self, actual_timestamp: datetime.datetime, actual_timestamp_ns: int) -> None:
        for group in self.__cache.message_groups.values():
            group.check_no_match_within_timeout(actual_timestamp=actual_timestamp,
                                                actual_timestamp_ns=actual_timestamp_ns,
                                                timeout=self.match_timeout)

    def clear_out_of_timeout(self, actual_timestamp: int) -> None:
        if self.autoremove_timeout is None:
            return
        elif isinstance(self.autoremove_timeout, datetime.datetime):
            autoremove_timestamp = self.autoremove_timeout.timestamp() * 1e9
            if autoremove_timestamp < actual_timestamp:
                self.autoremove_timeout += datetime.timedelta(days=1)
                self.clear_cache()
            return
        else:  # self.autoremove_timeout is int
            autoremove_timestamp = actual_timestamp - self.autoremove_timeout * 1e9
        for group in self.__cache.message_groups.values():
            group.clear_out_of_timeout(autoremove_timestamp)

    def stop(self) -> None:
        self.__cache.stop()



# Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import logging
from typing import Any, List

from th2_grpc_common.common_pb2 import EventStatus, FilterOperation
from th2_grpc_util.util_pb2 import ComparisonEntry, ComparisonEntryStatus, ComparisonEntryType

logger = logging.getLogger(__name__)


class EventType(str, enum.Enum):
    ROOT = 'ReconRoot'
    RULE = 'ReconRule'
    STATUS = 'ReconStatus'
    EVENT = 'ReconEvent'
    UNKNOWN = ''


class ComparatorUtils:

    @staticmethod
    def __get_result_count(comparison_result: ComparisonEntry, status: ComparisonEntryStatus) -> int:
        count = sum(ComparatorUtils.__get_result_count(sub_result, status)
                    for sub_result in comparison_result.fields.values())
        if status == comparison_result.status:
            count += 1

        return count

    @staticmethod
    def __get_status_type_by_val(failed: int, passed: int) -> ComparisonEntryStatus:
        if failed != 0:
            return ComparisonEntryStatus.FAILED  # type: ignore
        else:
            if passed != 0:
                return ComparisonEntryStatus.PASSED  # type: ignore
        return ComparisonEntryStatus.NA  # type: ignore

    @staticmethod
    def get_status_type(comparison_result: ComparisonEntry) -> ComparisonEntryStatus:
        failed = ComparatorUtils.__get_result_count(
            comparison_result, ComparisonEntryStatus.FAILED)  # type: ignore
        passed = ComparatorUtils.__get_result_count(
            comparison_result, ComparisonEntryStatus.PASSED)  # type: ignore
        return ComparatorUtils.__get_status_type_by_val(failed, passed)


class MessageComponent:

    def __init__(self, message: str) -> None:
        self.type = 'message'
        self.data = message


class TableComponent:

    def __init__(self, column_names: list) -> None:
        self.type = 'table'
        self.rows: List[dict] = []
        self.__column_names = column_names

    def add_row(self, *values: Any) -> None:
        self.rows.append({
            column_name: value for column_name, value
            in zip(self.__column_names, values)
        })


class VerificationComponent:

    def __init__(self, comparison_entry: ComparisonEntry) -> None:
        self.type = 'verification'
        self.fields = {
            field_name: VerificationComponent.VerificationEntry(entry)
            for field_name, entry in comparison_entry.fields.items()
        }
        self.status = EventStatus.FAILED \
            if ComparatorUtils.get_status_type(comparison_entry) == ComparisonEntryStatus.FAILED \
            else EventStatus.SUCCESS

    class VerificationEntry:

        def __init__(self, comparison_entry: ComparisonEntry) -> None:
            self.type = ComparisonEntryType.Name(comparison_entry.type).lower()
            self.status = ComparisonEntryStatus.Name(comparison_entry.status)
            self.expected = comparison_entry.first
            self.actual = comparison_entry.second
            self.key = comparison_entry.is_key
            self.operation = FilterOperation.Name(comparison_entry.operation)
            self.fields = {}

            if comparison_entry.type == ComparisonEntryType.COLLECTION:
                for field_name, entry in comparison_entry.fields.items():
                    self.fields[field_name] = VerificationComponent.VerificationEntry(entry)



# Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
from typing import Dict, Optional

from grpc import Server
from th2_check2_recon.configuration import ReconConfiguration
from th2_check2_recon.handler import GRPCHandler
from th2_check2_recon.reconcommon import MessageGroupDescription, ReconMessage
from th2_check2_recon.services import Cache, EventStore, MessageComparator
from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.message.message_router import MessageRouter
from th2_grpc_crawler_data_processor import crawler_data_processor_pb2_grpc

logger = logging.getLogger(__name__)


class Recon:

    def __init__(self,
                 event_router: EventBatchRouter,
                 message_router: MessageRouter,
                 custom_config: dict,
                 message_comparator: Optional[MessageComparator] = None,
                 grpc_server: Optional[Server] = None) -> None:
        logger.info('Recon initializing...')
        self.rules: list = []
        self._config = ReconConfiguration(**custom_config)
        self.__message_router: MessageRouter = message_router
        self.event_store = EventStore(recon_name=self._config.recon_name,
                                      event_router=event_router,
                                      event_batch_max_size=self._config.event_batch_max_size,
                                      event_batch_sending_interval=self._config.event_batch_send_interval)
        self.message_comparator: Optional[MessageComparator] = message_comparator
        self.grpc_server: Optional[Server] = grpc_server
        self.shared_message_groups: Dict[str, Cache.MessageGroup] = {}

    def start(self) -> None:
        try:
            logger.info('Recon is starting...')
            self.rules = self.__load_rules()
            self.__check_shared_message_groups()

            for rule in self.rules:
                self.__message_router.subscribe_all(rule.get_listener(), *rule.get_attributes())

            if self.grpc_server is not None:
                grpc_handler = GRPCHandler(self.rules, self._config.crawler_connection_configuration,
                                           self.event_store.recon_event_id)
                crawler_data_processor_pb2_grpc.add_DataProcessorServicer_to_server(grpc_handler, self.grpc_server)
                self.grpc_server.start()

            logger.info('Recon have started!')
        except Exception:
            logger.exception('Recon work interrupted')
            raise

    def stop(self) -> None:
        logger.info('Recon is trying to stop')
        try:
            self.__message_router.unsubscribe_all()
            for rule in self.rules:
                rule.stop()
            if self.message_comparator is not None:
                self.message_comparator.stop()
            self.event_store.stop()
        except Exception as e:
            logger.exception(f'Error while stopping Recon: {e}')
        finally:
            logger.info('Recon was stopped!')

    def __load_rules(self) -> list:
        logger.info('Try to load rules')
        rules_package = importlib.import_module(self._config.rules_package_path)
        loaded_rules = []
        for number, rule_config in enumerate(self._config.rules):
            if rule_config.enabled:
                module = importlib.import_module(rules_package.__name__ + '.' + rule_config.name)
                match_timeout = rule_config.match_timeout * 1_000_000_000 + rule_config.match_timeout_offset_ns
                loaded_rules.append(module.Rule(recon=self,
                                                cache_size=self._config.cache_size,
                                                match_timeout=match_timeout,
                                                match_all=rule_config.match_all,
                                                autoremove_timeout=rule_config.autoremove_timeout,
                                                configuration=rule_config.configuration,
                                                metric_number=number + 1))
        logger.info('Rules were loaded')
        return loaded_rules

    def __check_shared_message_groups(self) -> None:
        shared_group_descriptions_by_group_name: Dict[str, MessageGroupDescription] = {}

        for rule in self.rules:
            for group_name, group_description in rule.description_of_groups().items():
                if group_description.shared:
                    if group_name not in shared_group_descriptions_by_group_name:
                        shared_group_descriptions_by_group_name[group_name] = group_description
                    elif shared_group_descriptions_by_group_name[group_name] != group_description:
                        raise AttributeError(f'Shared group "{group_name}" should have the same '
                                             f'attributes in all rules')

    def put_shared_message(self, recon_message: ReconMessage) -> None:
        for rule in self.rules:
            if recon_message.group_name is None:
                raise AttributeError('Group name should be set in ReconMessage before putting it in shared group')
            message_group = recon_message.group_name
            rule_groups: Dict[str, MessageGroupDescription] = rule.description_of_groups()
            if message_group in rule_groups:
                if rule_groups[message_group].shared:
                    recon_message._shared = True
                else:
                    raise AttributeError(
                        f'Trying to put shared message in group that '
                        f'is not shared: {message_group}')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['th2_check2_recon']

package_data = \
{'': ['*']}

install_requires = \
['sortedcollections==2.1.0',
 'th2-common==3.10.0.dev3361836539',
 'th2-grpc-crawler-data-processor==0.4.0.dev3361134587',
 'th2-grpc-util==3.1.0.dev3360864447']

setup_kwargs = {
    'name': 'th2-check2-recon',
    'version': '4.0.0.dev3369120763',
    'description': 'Python library for creating th2-check2-recon applications.',
    'long_description': 'This repository is a library for creating th2-check2-recon applications.\n\n## Installation\n```\npip install th2-check2-recon\n```\nThis package can be found on [PyPI](https://pypi.org/project/th2-check2-recon/ "th2-check2-recon").\n',
    'author': 'TH2-devs',
    'author_email': 'th2-devs@exactprosystems.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/th2-net/th2-check2-recon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
from typing import Any, Dict, Optional, Union

from th2_grpc_common.common_pb2 import ConnectionID, Direction, MessageID


class ReconMessage:

    __slots__ = (
        'proto_message', 'group_name', 'hash', 'group_info', 'hash_info',
        '_is_matched', '_shared', '_timestamp', '_timestamp_ns', '_info',
        '_was_checked_no_match_within_timeout'
    )

    def __init__(self, proto_message: Dict[str, Any]) -> None:
        self.proto_message = proto_message
        self.group_name: Optional[str] = None
        self.hash = None

        self.group_info: dict = {}
        self.hash_info: dict = {}

        self._is_matched: bool = False
        self._shared: bool = False
        self._timestamp: Optional[datetime.datetime] = None
        self._timestamp_ns: Optional[int] = None
        self._info: Optional[str] = None
        self._was_checked_no_match_within_timeout: bool = False

    @property
    def timestamp(self) -> datetime.datetime:
        if self._timestamp is None:
            self._timestamp = ReconMessageUtils.get_timestamp(self.proto_message)
        return self._timestamp

    @property
    def timestamp_ns(self) -> int:
        if self._timestamp_ns is None:
            self._timestamp_ns = ReconMessageUtils.get_timestamp_ns(self.proto_message)
        return self._timestamp_ns

    @staticmethod
    def get_info(info_dict: dict) -> str:
        return '[' + ', '.join(f"'{key}': {value}" for key, value in info_dict.items()) + ']'

    @property
    def all_info(self) -> str:
        if self._info is None:
            result = f"'{self.proto_message['metadata']['message_type']} " \
                     f"'id='{ReconMessageUtils.str_message_id(self.proto_message)}'"
            self._info = result
        else:
            result = self._info
        if self.hash is not None:
            result += f" Hash='{self.hash}'"
        if self.group_name is not None:
            result += f" Group='{self.group_name}'"
        if len(self.hash_info) > 0:
            result += f' Hash{self.get_info(self.hash_info)}'
        if len(self.group_info) > 0:
            result += f' GroupID{self.get_info(self.group_info)}'
        return result


def _get_msg_timestamp(msg: ReconMessage) -> datetime.datetime:
    """Used instead of lambda in Rule.__check_and_store_event"""
    return msg.timestamp


class MessageGroupDescription:

    __slots__ = ('__single', '__multi', '__shared', '__ignore_no_match')

    def __init__(self,
                 single: bool = False,
                 multi: bool = False,
                 shared: bool = False,
                 ignore_no_match: bool = False):
        if single ^ multi:  # xor
            self.__single = single
            self.__multi = multi
        else:
            raise AttributeError('Group should be either single or multi')
        self.__shared = shared
        self.__ignore_no_match = ignore_no_match

    def __eq__(self, other: object) -> bool:
        properties = {'single', 'multi', 'shared', 'ignore_no_match'}
        if not isinstance(other, MessageGroupDescription):
            return NotImplemented
        return all(getattr(self, prop) == getattr(other, prop) for prop in properties)

    def __hash__(self) -> int:
        return hash((
            self.single, self.multi,
            self.shared, self.ignore_no_match
        ))

    @property
    def single(self) -> bool:
        return self.__single

    @property
    def multi(self) -> bool:
        return self.__multi

    @property
    def shared(self) -> bool:
        return self.__shared

    @property
    def ignore_no_match(self) -> bool:
        return self.__ignore_no_match


class ReconMessageUtils:
    """Some service methods for work with recon specifically and it's rules"""

    @staticmethod
    def get_timestamp(message: Dict[str, Any]) -> datetime.datetime:
        timestamp: datetime.datetime = message['metadata']['timestamp']
        return timestamp if timestamp is not None else datetime.datetime.min

    @staticmethod
    def get_timestamp_ns(message: Dict[str, Any]) -> int:
        timestamp = message['metadata']['timestamp']
        return int(timestamp.timestamp() * 10 ** 9) if timestamp is not None else 0

    @staticmethod
    def str_message_id(message: Dict[str, Any]) -> str:
        res = ''
        params = message['metadata']['session_alias'], message['metadata']['direction'], message['metadata']['sequence']
        for param in params:
            res += str(param) + ':' if param else 'None: '
        return res

    @staticmethod
    def get_value(message: ReconMessage, name: str, default: Any = '') -> Any:
        """return simple value of a given field if presented in the recon message."""
        return message.proto_message['fields'].get(name, default)

    @staticmethod
    def get_required_value(message: ReconMessage, name: str) -> Any:
        """return simple value of a given field."""
        return message.proto_message['fields'][name]

    @staticmethod
    def get_inner_value(message: ReconMessage, *names: Union[str, int], default: Any = '') -> Any:
        value = message.proto_message['fields']
        for name in names:
            if isinstance(name, int):
                if isinstance(value, list) and -len(value) <= name < len(value):
                    value = value[name]
                else:
                    value = None
            else:
                value = value.get(name)

            if value is None:
                return default
        return value

    @staticmethod
    def get_message_type(message: ReconMessage) -> str:
        return message.proto_message['metadata']['message_type']  # type: ignore

    @staticmethod
    def get_session_alias(message: ReconMessage) -> str:
        return message.proto_message['metadata']['session_alias']  # type: ignore

    @staticmethod
    def get_message_id(message: ReconMessage) -> MessageID:
        return MessageID(connection_id=ConnectionID(session_alias=message.proto_message['metadata']['session_alias'],
                                                    session_group=message.proto_message['metadata']['session_group']),
                         direction=getattr(Direction, message.proto_message['metadata']['direction']),
                         sequence=message.proto_message['metadata']['sequence'],
                         subsequence=message.proto_message['metadata']['subsequence'])

    @staticmethod
    def equal_values_check(message1: ReconMessage, message2: ReconMessage, tag: str) -> bool:
        """check that tag values are equal for both messages"""
        return ReconMessageUtils.get_value(message1, tag) == ReconMessageUtils.get_value(message2, tag)  # type: ignore

    @staticmethod
    def equal_if_not_empty_values_check(message1: ReconMessage, message2: ReconMessage, tag: str) -> bool:
        """check that tag values are equal (if they are not empty) for both messages"""
        value1 = ReconMessageUtils.get_value(message1, tag)
        value2 = ReconMessageUtils.get_value(message2, tag)
        if value1 != '' and value2 != '':
            return value1 == value2  # type: ignore
        return True

    @staticmethod
    def float_equal_values_check(message1: ReconMessage, message2: ReconMessage, tag: str) -> bool:
        """check that required float tag values are equal for both messages"""
        if ReconMessageUtils.get_value(message1, tag) != '' and ReconMessageUtils.get_value(message2, tag) != '':
            return float(ReconMessageUtils.get_value(message1, tag)) == \
                   float(ReconMessageUtils.get_value(message2, tag))
        return False

    @staticmethod
    def float_equal_if_not_empty_values_check(message1: ReconMessage, message2: ReconMessage, tag: str) -> bool:
        """check that float tag values (if they are not empty) are equal for both messages"""
        if ReconMessageUtils.get_value(message1, tag) != '' and ReconMessageUtils.get_value(message2, tag) != '':
            return float(ReconMessageUtils.get_value(message1, tag)) == \
                   float(ReconMessageUtils.get_value(message2, tag))
        return True

    @staticmethod
    def enum_value_check(message: ReconMessage, tag: str, enum_list: list) -> bool:
        return ReconMessageUtils.get_value(message, tag) in enum_list

    @staticmethod
    def not_zero_value_check(message: ReconMessage, tag: str) -> bool:
        """check that tag values isn't equal to zero for both messages"""
        val = ReconMessageUtils.get_value(message, tag)
        if val == '':
            return False
        return float(val) != 0

    @staticmethod
    def zero_value_check(message: ReconMessage, tag: str) -> bool:
        """check that tag value equals to zero"""
        val = ReconMessageUtils.get_value(message, tag)
        if val == '':
            return False
        return float(val) == 0
